/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.function.Function;
import java.util.function.IntFunction;
import java.util.function.IntPredicate;
import java.util.function.Predicate;
import java.util.function.UnaryOperator;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import org.apache.logging.log4j.CloseableThreadContext;
import org.apache.logging.log4j.ThreadContext;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import de.ozgcloud.aggregation.transformation.Transformation;
import de.ozgcloud.aggregation.transformation.TransformationException;
import de.ozgcloud.aggregation.transformation.TransformationService;
import de.ozgcloud.aggregation.transformation.VorgangMapper;
import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.aggregation.warehouse.WarehouseRepository;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgang;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQuery;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangService;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStub;
import de.ozgcloud.apilib.vorgang.Page;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@RequiredArgsConstructor
@Log4j2
@ConditionalOnProperty(prefix = "ozgcloud.command.line.runner", value = "enabled", havingValue = "true", matchIfMissing = true)
public class AggregationManagerRunner implements CommandLineRunner {

	private static final String MDC_EXECUTION = "execution";
	private static final String MDC_VORGANG = "vorgang";

	private static final Predicate<Batch> HAS_NEXT_BATCH = x -> !x.items.isEmpty();

	private final OzgCloudVorgangService vorgangService;
	private final AggregationManagerConfiguration config;

	private final TransformationProperties transformationProperties;

	private final TransformationService transformationService;
	private final WarehouseRepository repository;
	private final VorgangMapper vorgangMapper;

	@Override
	public void run(String... args) throws TransformationException {
		var identifier = transformationProperties.getIdentifier();
		var aggregationMappings = transformationProperties.getAggregationMappings();
		if (Objects.isNull(aggregationMappings) || aggregationMappings.isEmpty()) {
			runWithTransformation(transformationService.load(identifier, null));
		} else {
			aggregationMappings.stream()
					.map(aggregationMapping -> transformationService.load(identifier, aggregationMapping))
					.forEach(this::runWithTransformation);
		}
	}

	void runWithTransformation(Transformation transformation) {
		try (Execution execution = new Execution(transformation)) {
			ThreadContext.put(MDC_EXECUTION, execution.id.toString());
			loadVorgaengeIntoRepository(Stream.concat(
					extractBatchesOfVorgaengeFromDataSource(execution),
					extractBatchesOfDeletedVorgaengeFromDataSource(execution)));
		} finally {
			ThreadContext.remove(MDC_EXECUTION);
		}
	}

	void loadVorgaengeIntoRepository(Stream<Batch> batches) {
		batches.map(this::transformBatchToDocumentEntries).forEach(this::loadDocumentEntriesIntoRepository);
	}

	Stream<Batch> extractBatchesOfVorgaengeFromDataSource(Execution execution) {
		return extractBatchesFromDataSource(execution, this::getVorgaengeFromDataSource);
	}

	List<OzgCloudVorgang> getVorgaengeFromDataSource(Page page) {
		return vorgangService.find(buildFindAllQuery(), page).stream()
				.map(vorgangStub -> vorgangService.getById(vorgangStub.getId()))
				.toList();
	}

	OzgCloudVorgangQuery buildFindAllQuery() {
		return OzgCloudVorgangQuery.builder().build();
	}

	Stream<Batch> extractBatchesOfDeletedVorgaengeFromDataSource(Execution execution) {
		return extractBatchesFromDataSource(execution, getPagedDeletedVorgaenge(vorgangService.findDeleted()));
	}

	Function<Page, List<OzgCloudVorgang>> getPagedDeletedVorgaenge(Stream<OzgCloudVorgangStub> allDeletedVorgaenge) {
		var it = allDeletedVorgaenge.iterator();
		IntPredicate hasNextVorgangStub = ignored -> it.hasNext();
		IntFunction<OzgCloudVorgangStub> nextVorgangStub = ignored -> it.next();
		return page -> IntStream.range(page.getOffset(), page.getOffset() + page.getLimit())
				.takeWhile(hasNextVorgangStub)
				.mapToObj(nextVorgangStub)
				.map(vorgangMapper::fromVorgangStub)
				.toList();
	}

	Stream<Batch> extractBatchesFromDataSource(Execution execution, Function<Page, List<OzgCloudVorgang>> getFromDataSource) {
		var fetchSize = config.getFetchingBatchSize();
		var initialBatch = createBatch(execution, Page.builder().offset(0).limit(fetchSize).build(), getFromDataSource);
		UnaryOperator<Batch> nextBatch = x -> createBatch(
				execution,
				Page.builder().offset(x.page.getOffset() + fetchSize).limit(fetchSize).build(),
				getFromDataSource);
		return Stream.iterate(initialBatch, HAS_NEXT_BATCH, nextBatch).filter(x -> !x.items.isEmpty());
	}

	private Batch createBatch(Execution execution, Page page, Function<Page, List<OzgCloudVorgang>> getFromDataSource) {
		return new Batch(execution, createBatchUUID(), page, getFromDataSource.apply(page));
	}

	UUID createBatchUUID() {
		return UUID.randomUUID();
	}

	void loadDocumentEntriesIntoRepository(Stream<DocumentEntry> entries) {
		final List<DocumentEntry> documents = entries.toList();
		LOG.atDebug().log("store documents: {}", () -> documents.stream().map(DocumentEntry::getId).toList());
		repository.saveAll(documents);
	}

	Stream<DocumentEntry> transformBatchToDocumentEntries(Batch batch) {
		return batch.items.stream().map(vorgang -> transformWithinBatch(batch, vorgang)).filter(Objects::nonNull);
	}

	DocumentEntry transformWithinBatch(Batch batch, OzgCloudVorgang vorgang) {
		try (var instance = CloseableThreadContext.put(MDC_VORGANG, vorgang.getId().toString())) {
			return batch.execution.transformation.apply(vorgang);
		} catch (TransformationException e) {
			LOG.atError().withThrowable(e).log("failed to transform vorgang [{}] in batch [{}] of execution [{}]",
					vorgang::getId, () -> batch.id, () -> batch.execution.id);
			return null;
		}
	}

	@RequiredArgsConstructor
	protected static class Execution implements AutoCloseable {
		private final UUID id = UUID.randomUUID();
		@Getter
		private final Transformation transformation;

		@Override
		public void close() {
			// Nothing to do here
		}
	}

	record Batch(Execution execution, UUID id, Page page, List<OzgCloudVorgang> items) {
	}

}
