/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.ArgumentMatcher;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.AggregationManagerRunner.Batch;
import de.ozgcloud.aggregation.AggregationManagerRunner.Execution;
import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.aggregation.transformation.AggregationMappingTestFactory;
import de.ozgcloud.aggregation.transformation.Transformation;
import de.ozgcloud.aggregation.transformation.TransformationService;
import de.ozgcloud.aggregation.transformation.VorgangMapper;
import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.aggregation.warehouse.DocumentEntryTestFactory;
import de.ozgcloud.aggregation.warehouse.WarehouseRepository;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgang;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQuery;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangService;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStub;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStubTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import de.ozgcloud.apilib.vorgang.Page;

class AggregationManagerRunnerTest {

	@Mock
	private OzgCloudVorgangService vorgangService;
	@Mock
	private AggregationManagerConfiguration config;
	@Mock
	private TransformationProperties transformationProperties;
	@Mock
	private TransformationService transformationService;
	@Mock
	private WarehouseRepository repository;
	@Mock
	private VorgangMapper vorgangMapper;
	@Spy
	@InjectMocks
	private AggregationManagerRunner runner;

	@Nested
	class TestRun {
		private final String identifier = LoremIpsum.getInstance().getWords(1);

		@BeforeEach
		void mock() {
			when(transformationProperties.getIdentifier()).thenReturn(identifier);
			doNothing().when(runner).runWithTransformation(any());
		}

		@Test
		void shouldGetAggregationMappings() {
			runner.run();

			verify(transformationProperties).getAggregationMappings();
		}

		@Test
		void shouldGetIdentifier() {
			runner.run();

			verify(transformationProperties).getIdentifier();
		}

		@Nested
		class TestOnAggregationMappingConfigured {
			private final List<AggregationMapping> aggregationMappings = List.of(AggregationMappingTestFactory.create(),
					AggregationMappingTestFactory.create());

			@Mock
			private Transformation firstTransformation;
			@Mock
			private Transformation secondTransformation;

			@BeforeEach
			void mock() {
				when(transformationProperties.getAggregationMappings()).thenReturn(aggregationMappings);
				when(transformationService.load(any(), any())).thenReturn(firstTransformation, secondTransformation);
			}

			@Test
			void shouldLoadTransformationForEachAggregationMapping() {
				runner.run();

				aggregationMappings.forEach(mapping -> verify(transformationService).load(identifier, mapping));
			}

			@Test
			void shouldRunWithTransformationForEachTransformation() {
				runner.run();

				verify(runner).runWithTransformation(firstTransformation);
				verify(runner).runWithTransformation(secondTransformation);
			}
		}

		@Nested
		class TestOnAggregationMappingsNull {
			@Mock
			private Transformation transformation;

			@BeforeEach
			void mock() {
				when(transformationProperties.getAggregationMappings()).thenReturn(null);
				when(transformationService.load(any(), any())).thenReturn(transformation);
			}

			@Test
			void shouldLoadTransformation() {
				runner.run();

				verify(transformationService).load(identifier, null);
			}

			@Test
			void shouldCallRunWithTransformation() {
				runner.run();

				verify(runner).runWithTransformation(transformation);
			}
		}

		@Nested
		class TestOnAggregationMappingsEmpty {
			@Mock
			private Transformation transformation;

			@BeforeEach
			void mock() {
				when(transformationProperties.getAggregationMappings()).thenReturn(Collections.emptyList());
				when(transformationService.load(any(), any())).thenReturn(transformation);
			}

			@Test
			void shouldLoadTransformation() {
				runner.run();

				verify(transformationService).load(identifier, null);
			}

			@Test
			void shouldCallRunWithTransformation() {
				runner.run();

				verify(runner).runWithTransformation(transformation);
			}
		}
	}

	@Nested
	class TestRunWithTransformation {

		@Mock
		private Transformation transformation;
		private final ArgumentMatcher<Execution> hasTransformation = execution -> execution.getTransformation().equals(transformation);
		@Mock
		private Batch batchOfVorgaenge;
		@Mock
		private Batch batchOfDeletedVorgaenge;
		@Captor
		private ArgumentCaptor<Stream<Batch>> batchStreamCaptor;

		@BeforeEach
		void init() {
			doReturn(Stream.of(batchOfVorgaenge)).when(runner).extractBatchesOfVorgaengeFromDataSource(any());
			doReturn(Stream.of(batchOfDeletedVorgaenge)).when(runner).extractBatchesOfDeletedVorgaengeFromDataSource(any());
			doNothing().when(runner).loadVorgaengeIntoRepository(any());
		}

		@Test
		void shouldExtractBatchesOfVorgaengeFromDataSource() {
			runner.runWithTransformation(transformation);

			verify(runner).extractBatchesOfVorgaengeFromDataSource(argThat(hasTransformation));
		}

		@Test
		void shouldExtractBatchesOfDeletedVorgaengeFromDataSource() {
			runner.runWithTransformation(transformation);

			verify(runner).extractBatchesOfDeletedVorgaengeFromDataSource(argThat(hasTransformation));
		}

		@Test
		void shouldLoadVorgaengeIntoRepository() {
			runner.runWithTransformation(transformation);

			verify(runner).loadVorgaengeIntoRepository(batchStreamCaptor.capture());
			assertThat(batchStreamCaptor.getValue()).containsExactly(batchOfVorgaenge, batchOfDeletedVorgaenge);
		}
	}

	@Nested
	class TestLoadVorgaengeIntoRepository {

		@Mock
		private Execution execution;
		@Mock
		private Batch batch;
		private final DocumentEntry documentEntry = DocumentEntryTestFactory.create();
		@Captor
		private ArgumentCaptor<Stream<DocumentEntry>> documentEntriesCaptor;

		@BeforeEach
		void init() {
			doReturn(Stream.of(documentEntry)).when(runner).transformBatchToDocumentEntries(any());
			doNothing().when(runner).loadDocumentEntriesIntoRepository(any());
		}

		@Test
		void shouldTransform() {
			loadVorgaengeIntoRepository();

			verify(runner).transformBatchToDocumentEntries(batch);
		}

		@Test
		void shouldLoadIntoRepository() {
			loadVorgaengeIntoRepository();

			verify(runner).loadDocumentEntriesIntoRepository(documentEntriesCaptor.capture());
			assertThat(documentEntriesCaptor.getValue()).containsExactly(documentEntry);
		}

		private void loadVorgaengeIntoRepository() {
			runner.loadVorgaengeIntoRepository(Stream.of(batch));
		}
	}

	@Nested
	class TestExtractBatchesOfVorgaengeFromDataSource {

		@Mock
		private Execution execution;
		@Captor
		private ArgumentCaptor<Function<Page, List<OzgCloudVorgang>>> functionToRetrieveDataCaptor;
		@Mock
		private Batch batch;
		@Mock
		private Page page;

		@BeforeEach
		void init() {
			doReturn(Stream.of(batch)).when(runner).extractBatchesFromDataSource(any(), any());
		}

		@Test
		void shouldExtract() {
			runner.extractBatchesOfVorgaengeFromDataSource(execution);

			verify(runner).extractBatchesFromDataSource(eq(execution), any());
		}

		@Test
		void shouldExtractWithDataRetrievalFunction() {
			runner.extractBatchesOfVorgaengeFromDataSource(execution);

			verify(runner).extractBatchesFromDataSource(eq(execution), functionToRetrieveDataCaptor.capture());
			functionToRetrieveDataCaptor.getValue().apply(page);
			verify(runner).getVorgaengeFromDataSource(page);
		}

		@Test
		void shouldReturnExtractedBatches() {
			var extracted = runner.extractBatchesOfVorgaengeFromDataSource(execution);

			assertThat(extracted).containsExactly(batch);
		}
	}

	@Nested
	class TestGetVorgaengeFromDataSource {

		private final Page page = Page.builder().offset(10).limit(2).build();
		private final OzgCloudVorgangQuery query = OzgCloudVorgangQuery.builder().build();

		@BeforeEach
		void init() {
			doReturn(query).when(runner).buildFindAllQuery();
			when(vorgangService.find(any(), any())).thenReturn(List.of(OzgCloudVorgangStubTestFactory.create()));
			when(vorgangService.getById(any())).thenReturn(OzgCloudVorgangTestFactory.create());
		}

		@Test
		void shouldCallVorgangService() {
			getVorgaengeFromDataSource();

			verify(vorgangService).find(query, page);
		}

		@Test
		void shouldGetVorgangDetails() {
			getVorgaengeFromDataSource();

			verify(vorgangService).getById(OzgCloudVorgangTestFactory.ID);
		}

		@Test
		void shouldReturnVorgangDetails() {
			var vorgaenge = getVorgaengeFromDataSource();

			assertThat(vorgaenge).usingRecursiveFieldByFieldElementComparator().containsExactly(OzgCloudVorgangTestFactory.create());
		}

		private List<OzgCloudVorgang> getVorgaengeFromDataSource() {
			return runner.getVorgaengeFromDataSource(page);
		}
	}

	@Nested
	class TestExtractBatchesOfDeletedVorgaengeFromDataSource {

		@Mock
		private Execution execution;
		@Mock
		private Function<Page, List<OzgCloudVorgang>> functionToRetrieveData;
		private final List<OzgCloudVorgangStub> deletedVorgaenge = List.of(OzgCloudVorgangStubTestFactory.create());
		@Captor
		private ArgumentCaptor<Stream<OzgCloudVorgangStub>> deletedVorgaengeCaptor;
		@Mock
		private Batch batch;

		@BeforeEach
		void init() {
			when(vorgangService.findDeleted()).thenReturn(deletedVorgaenge.stream());
			doReturn(functionToRetrieveData).when(runner).getPagedDeletedVorgaenge(any());
			doReturn(Stream.of(batch)).when(runner).extractBatchesFromDataSource(any(), any());
		}

		@Test
		void shouldFindDeleted() {
			runner.extractBatchesOfDeletedVorgaengeFromDataSource(execution);

			verify(vorgangService).findDeleted();
		}

		@Test
		void shouldGetPagedDeletedVorgaenge() {
			runner.extractBatchesOfDeletedVorgaengeFromDataSource(execution);

			verify(runner).getPagedDeletedVorgaenge(deletedVorgaengeCaptor.capture());
			assertThat(deletedVorgaengeCaptor.getValue()).usingRecursiveFieldByFieldElementComparator().containsExactlyElementsOf(deletedVorgaenge);
		}

		@Test
		void shouldExtractWithDataRetrievalFunction() {
			runner.extractBatchesOfDeletedVorgaengeFromDataSource(execution);

			verify(runner).extractBatchesFromDataSource(execution, functionToRetrieveData);
		}

		@Test
		void shouldReturnExtractedBatches() {
			var extracted = runner.extractBatchesOfDeletedVorgaengeFromDataSource(execution);

			assertThat(extracted).containsExactly(batch);
		}
	}

	@Nested
	class TestGetPagedDeletedVorgaenge {

		private static final int PAGE_SIZE = 2;

		private final Stream<OzgCloudVorgangStub> vorgangStubs = Stream.generate(OzgCloudVorgangStubTestFactory::create).limit(PAGE_SIZE + 1);

		@Test
		void shouldReturnFirstFullPage() {
			var pagingFunction = runner.getPagedDeletedVorgaenge(vorgangStubs);

			var vorgaenge = getFirstPage(pagingFunction);

			assertThat(vorgaenge).hasSize(PAGE_SIZE);
		}

		@Test
		void shouldReturnSecondPageWithOneElement() {
			var pagingFunction = runner.getPagedDeletedVorgaenge(vorgangStubs);
			getFirstPage(pagingFunction);

			var vorgaenge = getSecondPage(pagingFunction);

			assertThat(vorgaenge).hasSize(1);
		}

		private List<OzgCloudVorgang> getFirstPage(Function<Page, List<OzgCloudVorgang>> pagingFunction) {
			return pagingFunction.apply(Page.builder().offset(0).limit(PAGE_SIZE).build());
		}

		private List<OzgCloudVorgang> getSecondPage(Function<Page, List<OzgCloudVorgang>> pagingFunction) {
			return pagingFunction.apply(Page.builder().offset(PAGE_SIZE).limit(PAGE_SIZE).build());
		}
	}

	@Nested
	class TestExtractBatchesFromDataSource {

		private static final int BATCH_SIZE = 2;
		private static final List<OzgCloudVorgang> VORGAENGE_1 = List.of(OzgCloudVorgangTestFactory.create(), OzgCloudVorgangTestFactory.create());
		private static final List<OzgCloudVorgang> VORGAENGE_2 = List.of(OzgCloudVorgangTestFactory.create());
		private static final UUID BATCH_1_UUID = UUID.randomUUID();
		private static final UUID BATCH_2_UUID = UUID.randomUUID();

		@Mock
		private Execution execution;
		@Mock
		private Function<Page, List<OzgCloudVorgang>> functionToRetrieveData;

		@BeforeEach
		void init() {
			when(config.getFetchingBatchSize()).thenReturn(BATCH_SIZE);
			when(functionToRetrieveData.apply(any())).thenReturn(VORGAENGE_1, VORGAENGE_2, List.of());
			when(runner.createBatchUUID()).thenReturn(BATCH_1_UUID, BATCH_2_UUID);
		}

		@Test
		void shouldCallFunctionToRetrieveDataForFirstPage() {
			extractBatchesFromDataSource();

			verify(functionToRetrieveData).apply(argThat(hasExpectedOffsetAndConfiguredLimit(0)));
		}

		@Test
		void shouldCallFunctionToRetrieveDataForSecondPage() {
			extractBatchesFromDataSource().toList();

			verify(functionToRetrieveData).apply(argThat(hasExpectedOffsetAndConfiguredLimit(BATCH_SIZE)));
		}

		@Test
		void shouldCallFunctionToRetrieveDataForThirdPage() {
			extractBatchesFromDataSource().toList();

			verify(functionToRetrieveData).apply(argThat(hasExpectedOffsetAndConfiguredLimit(BATCH_SIZE * 2)));
		}

		@Test
		void shouldReturnBatches() {
			var batches = extractBatchesFromDataSource().toList();

			assertThat(batches).hasSize(2).usingRecursiveFieldByFieldElementComparator().containsExactly(
					new Batch(execution, BATCH_1_UUID, Page.builder().offset(0).limit(BATCH_SIZE).build(), VORGAENGE_1),
					new Batch(execution, BATCH_2_UUID, Page.builder().offset(BATCH_SIZE).limit(BATCH_SIZE).build(), VORGAENGE_2)
			);
		}

		private ArgumentMatcher<Page> hasExpectedOffsetAndConfiguredLimit(int expectedOffset) {
			return page -> page.getOffset() == expectedOffset && page.getLimit() == BATCH_SIZE;
		}

		private Stream<Batch> extractBatchesFromDataSource() {
			return runner.extractBatchesFromDataSource(execution, functionToRetrieveData);
		}
	}
}
