/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.transformation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.schibsted.spt.data.jslt.Expression;
import com.schibsted.spt.data.jslt.impl.DotExpression;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.transformation.AggregationMapping.FieldMapping;

class JSLTransformationServiceTest {

	@InjectMocks
	@Spy
	private JSLTransformationService service;
	@Mock
	private ObjectMapper objectMapper;
	@Mock
	private VorgangMapper vorgangMapper;

	@Nested
	class TestLoad {

		private final String identifier = LoremIpsum.getInstance().getWords(1);
		private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();
		private final Map<String, String> mapping = FieldMappingTestFactory.createAsMap();

		@Mock
		private Expression expression;
		private Optional<DotExpression> identifierExpression = Optional.of(DotExpressionTestFactory.create());

		@Nested
		class TestOnMappingNotNull {
			@BeforeEach
			void mock() {
				doReturn(mapping).when(service).transformToMap(any());
				doReturn(expression).when(service).createExpression(any());
				doReturn(identifierExpression).when(service).transformToDotExpression(any());
			}

			@Test
			void shouldTransformToMap() {
				load();

				verify(service).transformToMap(List.of(AggregationMappingTestFactory.MAPPING));
			}

			@Test
			void shouldCreateExpression() {
				load();

				verify(service).createExpression(mapping);
			}

			@Test
			void shouldReturnJSLTransformation() {
				var expectedTransformation = JSLTransformation.builder()
						.objectMapper(objectMapper)
						.vorgangMapper(vorgangMapper)
						.identifier(identifierExpression)
						.script(expression)
						.build();

				var transformation = load();

				assertThat(transformation).usingRecursiveComparison().isEqualTo(expectedTransformation);
			}

			private Transformation load() {
				return service.load(identifier, aggregationMapping);
			}
		}

		@Nested
		class TestOnMappingNull {
			@BeforeEach
			void mock() {
				doReturn(mapping).when(service).transformToMap(any());
				doReturn(expression).when(service).createExpression(any());
				doReturn(identifierExpression).when(service).transformToDotExpression(any());
			}

			@Test
			void shouldTransformToMap() {
				load();

				verify(service).transformToMap(Collections.emptyList());
			}

			@Test
			void shouldCreateExpression() {
				load();

				verify(service).createExpression(mapping);
			}

			@Test
			void shouldReturnJSLTransformation() {
				var expectedTransformation = JSLTransformation.builder()
						.objectMapper(objectMapper)
						.vorgangMapper(vorgangMapper)
						.identifier(identifierExpression)
						.script(expression)
						.build();

				var transformation = load();

				assertThat(transformation).usingRecursiveComparison().isEqualTo(expectedTransformation);
			}

			private Transformation load() {
				return service.load(identifier, null);
			}
		}
	}

	@Nested
	class TestTransformToMap {

		private static final String SECOND_SOURCE = LoremIpsum.getInstance().getWords(1);
		private static final String SECOND_TARGET = LoremIpsum.getInstance().getWords(1);
		private final List<FieldMapping> mappings = List.of(
				FieldMappingTestFactory.create(),
				FieldMappingTestFactory.createBuilder().sourcePath(SECOND_SOURCE).targetPath(SECOND_TARGET).build());

		@Test
		void shouldTransformToMap() {
			var expectedMap = Map.of(FieldMappingTestFactory.TARGET_PATH, FieldMappingTestFactory.SOURCE_PATH, SECOND_TARGET, SECOND_SOURCE);

			var map = service.transformToMap(mappings);

			assertThat(map).containsExactlyInAnyOrderEntriesOf(expectedMap);
		}
	}
}
