/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.transformation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.TextNode;
import com.schibsted.spt.data.jslt.Expression;
import com.schibsted.spt.data.jslt.impl.ExpressionNode;
import com.schibsted.spt.data.jslt.impl.Scope;

import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.aggregation.warehouse.DocumentEntry.DocumentEntryBuilder;
import de.ozgcloud.aggregation.warehouse.DocumentEntryTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgang;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import lombok.SneakyThrows;

class JSLTransformationTest {

	@Mock
	private ObjectMapper objectMapper;
	@Mock
	private Expression script;
	@Mock
	private ExpressionNode identifier;
	@Mock
	private VorgangMapper vorgangMapper;

	private JSLTransformation transformation;

	@Nested
	class TestApply {

		private final OzgCloudVorgang vorgang = OzgCloudVorgangTestFactory.create();
		private DocumentEntryBuilder documentEntryBuilder = DocumentEntryTestFactory.createBuilder().payload(null);

		@Mock
		private JsonNode vorgangTree;

		@Captor
		private ArgumentCaptor<DocumentEntry.DocumentEntryBuilder> documentEntryBuilderCaptor;

		@BeforeEach
		void mock() {
			transformation = spy(JSLTransformation.builder().objectMapper(objectMapper).vorgangMapper(vorgangMapper)
					.identifier(Optional.of(identifier)).script(script).build());
			when(objectMapper.valueToTree(vorgang)).thenReturn(vorgangTree);
			when(vorgangMapper.toDocumentEntryBuilder(any(), any())).thenReturn(documentEntryBuilder);
			doReturn(documentEntryBuilder).when(transformation).overrideId(any(), any());
			when(script.apply(any())).thenReturn(DocumentEntryTestFactory.PAYLOAD);
		}

		@Test
		void shouldCallVorgangMapper() {
			transformation.apply(vorgang);

			verify(vorgangMapper).toDocumentEntryBuilder(eq(vorgang), documentEntryBuilderCaptor.capture());
			assertThat(documentEntryBuilderCaptor.getValue()).usingRecursiveComparison().isEqualTo(DocumentEntry.builder());
		}

		@Test
		void shouldMapVorgangToObjectTree() {
			transformation.apply(vorgang);

			verify(objectMapper).valueToTree(vorgang);
		}

		@Test
		void shouldAplyScriptToVorgangTree() {
			transformation.apply(vorgang);

			verify(script).apply(vorgangTree);
		}

		@Test
		void shouldCallOverrideId() {
			transformation.apply(vorgang);

			verify(transformation).overrideId(vorgangTree, documentEntryBuilder);
		}

		@Test
		void shouldReturnDocumentEntry() {
			var entry = transformation.apply(vorgang);

			assertThat(entry).usingRecursiveComparison().isEqualTo(DocumentEntryTestFactory.create());
		}
	}

	@Nested
	class TestOverrideId {

		@Nested
		class TestOnIdentifierDefinitionGiven {

			private static final String ID = UUID.randomUUID().toString();
			private DocumentEntryBuilder documentEntryBuilder = DocumentEntryTestFactory.createBuilder().payload(null);
			@Mock
			private JsonNode vorgangTree;
			private Scope root;

			@BeforeEach
			@SneakyThrows
			void setUpTransformationClass() {
				transformation = spy(JSLTransformation.builder().objectMapper(objectMapper).vorgangMapper(vorgangMapper)
						.identifier(Optional.of(identifier)).script(script).build());
				var rootField = JSLTransformation.class.getDeclaredField("root");
				rootField.setAccessible(true);
				root = (Scope) rootField.get(transformation);
				when(identifier.apply(root, vorgangTree)).thenReturn(new TextNode(ID));
			}

			@Test
			void shouldApplyIdentifierDefinition() {
				transformation.overrideId(vorgangTree, documentEntryBuilder);

				verify(identifier).apply(root, vorgangTree);
			}

			@Test
			void shouldSetId() {
				when(identifier.apply(root, vorgangTree)).thenReturn(new TextNode(ID));

				var result = transformation.overrideId(vorgangTree, documentEntryBuilder);

				assertThat(result).usingRecursiveComparison().isEqualTo(DocumentEntryTestFactory.createBuilder().payload(null).id(ID));
			}
		}

		@Nested
		class TestOnIdentifierDefinitionNotGiven {

			private DocumentEntryBuilder documentEntryBuilder = DocumentEntryTestFactory.createBuilder().payload(null);
			@Mock
			private JsonNode vorgangTree;

			@BeforeEach
			@SneakyThrows
			void setUpTransformationClass() {
				transformation = spy(JSLTransformation.builder().objectMapper(objectMapper).vorgangMapper(vorgangMapper).script(script).build());
			}

			@Test
			void shouldReturnBuilder() {
				var result = transformation.overrideId(vorgangTree, documentEntryBuilder);

				assertThat(result).usingRecursiveComparison().isEqualTo(DocumentEntryTestFactory.createBuilder().payload(null));
			}
		}
	}
}
